"""
CLI interface and main logic for the web novel scraper.
"""
import argparse
import asyncio
import json
import os
import sys
from typing import List, Dict, Any, Optional, Tuple

import httpx
from bs4 import BeautifulSoup
from playwright.async_api import async_playwright
from simple_term_menu import TerminalMenu
from prompt_toolkit import PromptSession
from prompt_toolkit.completion import Completer, Completion, ThreadedCompleter
from loguru import logger
from rich.console import Console
from rich.table import Table
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn, TaskProgressColumn
from rich.panel import Panel

from . import tao_so_do_cay
from .scraper_core import lay_thong_tin_truyen, scrape_chapters
from .exporter import (
    tao_file_epub, tao_file_pdf, tao_file_html, tao_file_md, tao_file_txt, tao_file_mp3,
    ContentList, ChaptersData
)
from .utils import (
    sanitize_filename, create_folders_from_tree, normalize_vietnamese_url, 
    get_token_from_state, HEADERS, configure_logger
)
from .models import StoryInfo, ContentItem
from .session_manager import load_session, save_session, capture_session

SESSION_FILE = ".vvr_session.json"
BASE_URL = "https://valvrareteam.net"
console = Console()


class NovelCompleter(Completer):
    """Completer for live novel search using the Valvrare Team API."""
    def __init__(self, token: Optional[str] = None):
        self.token = token

    def get_completions(self, document, complete_event):
        text = document.text.strip()
        if len(text) < 3:
            return

        try:
            with httpx.Client(timeout=3.0) as client:
                headers = {
                    "User-Agent": HEADERS["User-Agent"],
                    "Origin": BASE_URL,
                    "Referer": f"{BASE_URL}/",
                    "Accept": "application/json, text/plain, */*"
                }
                if self.token:
                    headers["Authorization"] = f"Bearer {self.token}"
                
                url = f"https://val-ssr-2kzit.ondigitalocean.app/api/novels/search?title={text}"
                response = client.get(url, headers=headers)
                
                if response.status_code == 200:
                    results = response.json()
                    for item in results:
                        title = item.get('title', '').strip()
                        author = item.get('author', 'Unknown')
                        _id = item.get('_id', '')
                        status = item.get('status', 'Unknown')
                        total = item.get('totalChapters', 0)
                        
                        if title and _id:
                            slug = normalize_vietnamese_url(title) + "-" + _id[-8:]
                            meta = f"{author} | {status} | {total} ch"
                            yield Completion(
                                slug, 
                                start_position=-len(document.text), 
                                display=title, 
                                display_meta=meta
                            )
        except Exception:
            pass


class InteractiveUI:
    """Handles all interactive terminal elements."""
    
    @staticmethod
    def show_menu(items: List[str], title: str, multi_select: bool = False) -> Any:
        menu = TerminalMenu(
            items, 
            title=f" {title} ", 
            menu_cursor_style=("fg_cyan", "bold"), 
            menu_highlight_style=("bg_cyan", "fg_black"),
            multi_select=multi_select,
            show_multi_select_hint=multi_select
        )
        return menu.show()

    @staticmethod
    async def get_novel_name_interactive(token: Optional[str] = None) -> str:
        console.print("[yellow]Nhập tên truyện hoặc tìm kiếm (tối thiểu 3 ký tự để hiện gợi ý)...[/yellow]")
        try:
            session = PromptSession()
            name = await session.prompt_async(
                "Tên truyện: ",
                completer=ThreadedCompleter(NovelCompleter(token=token)),
                complete_while_typing=True
            )
            return name.strip()
        except (KeyboardInterrupt, Exception) as e:
            if isinstance(e, KeyboardInterrupt): raise
            return input("Nhập tên truyện bạn muốn tải: ").strip()

    @staticmethod
    def display_story_summary(info: StoryInfo):
        table = Table(title="[bold cyan]Thông tin truyện[/bold cyan]", show_header=True, header_style="bold magenta", border_style="bright_blue")
        table.add_column("Trường", style="cyan", width=15)
        table.add_column("Giá trị", style="white")
        
        table.add_row("Tiêu đề", info.title)
        table.add_row("Tác giả", info.author)
        table.add_row("Thể loại", ", ".join(info.genres) if info.genres else "N/A")
        desc = info.description
        if len(desc) > 200: desc = desc[:197] + "..."
        table.add_row("Mô tả", desc)
        
        console.print(table)


class ValvrareScraperCLI:
    """Main Orchestrator for the Valvrare Team Scraper CLI."""

    def __init__(self):
        self.args = self._parse_arguments()
        self.is_cli_mode = len(sys.argv) > 1
        self.session_state = None
        self.cookies = {}
        self.token = None
        self.skipped_urls = []
        self.output_folder = ""
        
        # Configure logging
        configure_logger(self.args.verbose)

    def _parse_arguments(self) -> argparse.Namespace:
        parser = argparse.ArgumentParser(
            description="Tải truyện từ Valvrare Team dưới dạng PDF, EPUB, và các định dạng khác.",
            formatter_class=argparse.RawTextHelpFormatter
        )
        parser.add_argument('ten_truyen', nargs='*', help="Tên truyện cần tải (slug). Hoặc dùng 'web' để mở giao diện.")
        parser.add_argument('-o', '--output', dest='output_folder', help="Thư mục đầu ra.")
        parser.add_argument('-f', '--format', nargs='+', default=['EPUB'], 
                           choices=['PDF', 'EPUB', 'HTML', 'MD', 'TXT'], help="Định dạng file.")
        parser.add_argument('-g', '--gop', default='rieng', choices=['rieng', 'volume', 'tatca'], 
                           help="Cách gộp file (rieng/volume/tatca).")
        parser.add_argument('--khong-minh-hoa', action='store_true', help="Bỏ qua minh họa.")
        parser.add_argument('--font', default='DejaVuSans', choices=['NotoSerif', 'DejaVuSans'], 
                           help="Font cho PDF.")
        parser.add_argument('-t', '--tasks', type=int, default=5, help="Số lượng tác vụ song song.")
        parser.add_argument('--login', action='store_true', help="Đăng nhập thủ công.")
        parser.add_argument('--refresh-session', action='store_true', help="Xóa session cũ.")
        parser.add_argument('--verbose', action='store_true', help="Hiển thị log chi tiết.")
        
        # Web-specific arguments (moved to top-level)
        parser.add_argument('--host', default='127.0.0.1', help='Host cho web server.')
        parser.add_argument('--port', type=int, default=8000, help='Port cho web server.')
        parser.add_argument('--workers', type=int, default=1, help='Số lượng novel tải song song (chế độ web).')
        parser.add_argument('--no-browser', action='store_true', help='Không tự động mở trình duyệt.')

        selection_group = parser.add_mutually_exclusive_group()
        selection_group.add_argument('--all', action='store_true', help="Tải tất cả.")
        selection_group.add_argument('--volumes', nargs='+', type=int, help="Tải các tập cụ thể.")
        selection_group.add_argument('--chapters', nargs='+', type=int, help="Tải các chương cụ thể.")
        
        return parser.parse_args()

    async def setup_session(self):
        """Handles session loading, login, and token extraction."""
        self.session_state = load_session(SESSION_FILE)
        
        if self.args.refresh_session and os.path.exists(SESSION_FILE):
            os.remove(SESSION_FILE)
            self.session_state = None

        if self.args.login or not self.session_state:
            do_login = self.args.login
            if not do_login and not self.is_cli_mode:
                choice = input("\nKhông tìm thấy session. Bạn có muốn đăng nhập không? (Y/n): ").strip().lower()
                do_login = not choice or choice in ['y', 'yes']
            
            if do_login:
                logger.info("Đang khởi tạo trình duyệt để lấy session...")
                self.session_state = await capture_session(f"{BASE_URL}/")
                save_session(self.session_state, SESSION_FILE)
                logger.success("Session đã được lưu.")

        self.token = get_token_from_state(self.session_state)
        if self.session_state and 'cookies' in self.session_state:
            for c in self.session_state['cookies']:
                self.cookies[c['name']] = c['value']

        if self.args.verbose:
            self._print_debug_info()

    def _print_debug_info(self):
        logger.debug(f"Số lượng cookies: {len(self.session_state.get('cookies', [])) if self.session_state else 0}")
        if self.token: logger.debug(f"Token (JWT): {self.token[:20]}...")
        logger.debug(f"Headers: {HEADERS['User-Agent']}")

    async def resolve_story_url(self, name_raw: str) -> Optional[str]:
        """Finds the story URL from sitemap."""
        normalized = normalize_vietnamese_url(name_raw)
        sitemap_url = f"{BASE_URL}/sitemap.xml"
        
        async with httpx.AsyncClient(headers=HEADERS, cookies=self.cookies) as client:
            try:
                response = await client.get(sitemap_url)
                response.raise_for_status()
                soup = BeautifulSoup(response.content, "lxml-xml")
                for loc in soup.find_all("loc"):
                    url = loc.text
                    if normalized in url and "/chuong" not in url:
                        return url
            except Exception as e:
                logger.error(f"Lỗi khi truy cập sitemap: {e}")
        return None

    def filter_chapters(self, chapter_data: List[Dict]) -> List[Dict]:
        """Applies filters like excluding illustrations/empty volumes."""
        if self.is_cli_mode:
            skip_minh_hoa = self.args.khong_minh_hoa
        else:
            choice = input("Bạn có muốn bỏ qua minh họa và chương lỗi? (Y/n): ").strip().lower()
            skip_minh_hoa = not choice or choice in ["y", "yes"]

        if skip_minh_hoa:
            return [vol for vol in chapter_data if vol['chapters']]
        return chapter_data

    def select_chapters_to_download(self, chapter_data: List[Dict]) -> List[Dict]:
        """Handles chapter selection logic for both CLI and Interactive modes."""
        selected = []
        if self.is_cli_mode:
            if self.args.volumes:
                for idx in self.args.volumes:
                    if 0 < idx <= len(chapter_data):
                        selected.extend(chapter_data[idx-1]['chapters'])
            elif self.args.chapters:
                flat = [c for v in chapter_data for c in v['chapters']]
                for idx in self.args.chapters:
                    if 0 < idx <= len(flat):
                        selected.append(flat[idx-1])
            else: # All
                selected = [c for v in chapter_data for c in v['chapters']]
        else:
            menu_idx = InteractiveUI.show_menu(
                ["Tải xuống tất cả", "Chọn tập để tải", "Chọn chương để tải"],
                "Tùy chọn tải xuống"
            )
            if menu_idx == 0:
                selected = [c for v in chapter_data for c in v['chapters']]
            elif menu_idx == 1:
                vols = [v['volume'] for v in chapter_data]
                v_idxs = InteractiveUI.show_menu(vols, "Chọn tập", multi_select=True)
                if v_idxs:
                    for i in v_idxs: selected.extend(chapter_data[i]['chapters'])
            elif menu_idx == 2:
                all_chaps = [(f"{v['volume']}: {c['title']}", c) for v in chapter_data for c in v['chapters']]
                c_idxs = InteractiveUI.show_menu([i[0] for i in all_chaps], "Chọn chương", multi_select=True)
                if c_idxs:
                    for i in c_idxs: selected.append(all_chaps[i][1])
        return selected

    async def run(self):
        """Main execution flow."""
        # Handle 'web' command as a positional argument
        if self.args.ten_truyen and self.args.ten_truyen[0] == 'web':
            from .web import run_web_server
            import webbrowser
            
            url = f"http://{self.args.host}:{self.args.port}"
            if not self.args.no_browser:
                webbrowser.open(url)
            
            await run_web_server(host=self.args.host, port=self.args.port, num_workers=self.args.workers)
            return

        await self.setup_session()

        # 1. Get Novel Names
        if self.is_cli_mode:
            names_raw = self.args.ten_truyen
            if not names_raw:
                # Should not happen if is_cli_mode is correctly handled
                # but if user passed other flags but no slug?
                logger.error("Tên truyện là bắt buộc ở chế độ CLI.")
                return
            
            total = len(names_raw)
            for i, name in enumerate(names_raw):
                if total > 1:
                    console.print(Panel(f"[bold cyan]Đang xử lý truyện {i+1}/{total}: {name}[/bold cyan]"))
                try:
                    await self.process_novel(name)
                except Exception as e:
                    logger.error(f"Lỗi khi xử lý '{name}': {e}")
                    if self.args.verbose:
                        import traceback
                        logger.error(traceback.format_exc())
        else:
            name_raw = await InteractiveUI.get_novel_name_interactive(self.token)
            if name_raw:
                await self.process_novel(name_raw)

    async def process_novel(self, name_raw: str):
        """Process a single novel download."""
        self.skipped_urls = []
        # 2. Resolve URL and Info
        story_url = await self.resolve_story_url(name_raw)
        if not story_url:
            logger.error(f"Không tìm thấy truyện '{name_raw}'.")
            return

        # Folder name is based on the slug from the URL
        relative_path = story_url.split(f"{BASE_URL}/")[-1]
        self.output_folder = self.args.output_folder or sanitize_filename(relative_path.split("/")[-1])
        os.makedirs(self.output_folder, exist_ok=True)

        async with httpx.AsyncClient(headers=HEADERS, cookies=self.cookies) as client:
            story_info = await lay_thong_tin_truyen(client, relative_path, verbose=self.args.verbose)
            if not self.is_cli_mode or self.args.verbose:
                InteractiveUI.display_story_summary(story_info)

        # 3. Load Chapter List
        logger.info(f"Đang lấy danh sách chương cho '{story_info.title}'...")
        await tao_so_do_cay.get_chapter_tree_list(story_url, output_file="chapter_list.json", session_state=self.session_state)
        with open("chapter_list.json", "r", encoding="utf-8") as f:
            chapter_data = json.load(f)

        # 4. Filter and Select
        chapter_data = self.filter_chapters(chapter_data)
        if not chapter_data: return
        
        selected_chaps = self.select_chapters_to_download(chapter_data)
        if not selected_chaps: return

        # 5. Export Config
        export_config = await self._get_export_config(story_url)
        
        # 6. Scrape
        urls = [f"{BASE_URL}{c['url']}" for c in selected_chaps]
        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=True)
            
            with Progress(
                SpinnerColumn(),
                TextColumn("[progress.description]{task.description}"),
                BarColumn(),
                TaskProgressColumn(),
                console=console,
                transient=True
            ) as progress:
                scrape_task = progress.add_task(f"[green]Đang tải {len(urls)} chương...", total=len(urls))
                
                scraped = await scrape_chapters(
                    browser, urls, export_config['tasks'], 
                    skipped_urls=self.skipped_urls, 
                    session_state=self.session_state, 
                    verbose=self.args.verbose,
                    token=self.token
                )
                progress.update(scrape_task, completed=len(urls))
                
            await browser.close()

        # 7. Generate Files
        logger.info(f"Bắt đầu tạo file cho '{story_info.title}'...")
        await self._generate_files(chapter_data, selected_chaps, scraped, story_info, export_config)
        self._cleanup()

    async def _get_export_config(self, story_url: str) -> Dict:
        if self.is_cli_mode:
            gop_map = {'rieng': 1, 'volume': 2, 'tatca': 0}
            return {
                'mode_idx': gop_map[self.args.gop],
                'formats': [f.upper() for f in self.args.format],
                'font': self.args.font,
                'tasks': self.args.tasks
            }
        
        mode_idx = InteractiveUI.show_menu(
            ["Gộp tất cả (mặc định)", "Xuất riêng từng chương", "Gộp theo Volume"],
            "Chọn cách thức xuất file"
        )
        if mode_idx in [1, 2]:
            tree_path = os.path.join(self.output_folder, "tree_map.txt")
            await tao_so_do_cay.get_chapter_tree_folder(url=story_url, output_file=tree_path, cookies=self.cookies)
            create_folders_from_tree(tree_path, self.output_folder)

        format_items = ["PDF", "EPUB", "HTML", "Markdown (.md)", "Text (.txt)", "MP3 (Audiobook)"]
        f_idxs = InteractiveUI.show_menu(format_items, "Chọn định dạng file", multi_select=True)
        if not f_idxs: return {}
        
        mapping = {"PDF": "PDF", "EPUB": "EPUB", "HTML": "HTML", "Markdown (.md)": "MD", "Text (.txt)": "TXT", "MP3 (Audiobook)": "MP3"}
        formats = [mapping[format_items[i]] for i in f_idxs]
        
        font = 'DejaVuSans'
        if "PDF" in formats:
            f_choice = input("Chọn font PDF (1. Noto Serif, 2. DejaVu Sans): ").strip()
            if f_choice == '1': font = 'NotoSerif'

        tasks_in = input("Số lượng tác vụ song song (mặc định 5): ")
        tasks = int(tasks_in) if tasks_in.isdigit() and int(tasks_in) > 0 else 5
        
        return {'mode_idx': mode_idx, 'formats': formats, 'font': font, 'tasks': tasks}

    async def _generate_files(self, chapter_data, selected_chaps, scraped, story_info, config):
        # Map URL to metadata
        url_to_vol = {c['url']: v['volume'] for v in chapter_data for c in v['chapters']}
        url_to_title = {f"{BASE_URL}{c['url']}": c['title'] for c in selected_chaps}
        
        mode = config['mode_idx']
        
        if mode == 1: # Rieng
            for url, content in scraped.items():
                rel_url = url.replace(BASE_URL, "")
                vol_name = url_to_vol.get(rel_url, "Unknown")
                folder = os.path.join(self.output_folder, sanitize_filename(vol_name))
                os.makedirs(folder, exist_ok=True)
                title = url_to_title.get(url, "Chapter")
                await self._write_to_formats(folder, title, content, story_info, config, [{'title': title, 'content': content}])

        elif mode == 2: # Volume
            vol_map = {}
            for url, content in scraped.items():
                v = url_to_vol.get(url.replace(BASE_URL, ""), "Unknown")
                if v not in vol_map: vol_map[v] = []
                vol_map[v].append({'title': url_to_title[url], 'content': content})
            
            for vol_name, chapters in vol_map.items():
                folder = os.path.join(self.output_folder, sanitize_filename(vol_name))
                os.makedirs(folder, exist_ok=True)
                full_content = [item for c in chapters for item in c['content']]
                await self._write_to_formats(folder, vol_name, full_content, story_info, config, chapters)

        else: # Tat ca
            full_structure = []
            full_flat = []
            for v_info in chapter_data:
                v_chaps = []
                for c_entry in v_info['chapters']:
                    f_url = f"{BASE_URL}{c_entry['url']}"
                    if f_url in scraped:
                        v_chaps.append({'title': c_entry['title'], 'content': scraped[f_url]})
                        full_flat.extend(scraped[f_url])
                if v_chaps:
                    full_structure.append({'volume': v_info['volume'], 'chapters': v_chaps})
            
            # Use real title for the filename
            await self._write_to_formats(self.output_folder, story_info.title, full_flat, story_info, config, full_structure)

    async def _write_to_formats(self, folder, title, content, info, config, structure):
        for fmt in config['formats']:
            fname = sanitize_filename(title)
            fpath = os.path.join(folder, f"{fname}.{fmt.lower()}")
            if fmt == "PDF": await tao_file_pdf(content, fpath, title, config['font'])
            elif fmt == "EPUB": await tao_file_epub(fpath, title, info.author, structure, info.description, info.cover_path, info.genres)
            elif fmt == "HTML": await tao_file_html(content, fpath, title)
            elif fmt == "MD": await tao_file_md(content, fpath, title)
            elif fmt == "TXT": await tao_file_txt(content, fpath, title)
            elif fmt == "MP3": await tao_file_mp3(content, fpath, title)

    def _cleanup(self):
        logger.success("--- HOÀN TẤT ---")
        if self.skipped_urls:
            log_path = os.path.join(self.output_folder, "cac_chuong_da_bo_qua.txt")
            logger.warning(f"{len(self.skipped_urls)} chương bị lỗi. Xem tại: {log_path}")
            with open(log_path, "w", encoding="utf-8") as f:
                for url in self.skipped_urls: f.write(f"{url}\n")


def main():
    """Entry point for the CLI."""
    cli = ValvrareScraperCLI()
    try:
        asyncio.run(cli.run())
    except KeyboardInterrupt:
        console.print("\n[bold red]Chương trình bị dừng bởi người dùng.[/bold red]")
    finally:
        # Cleanup temporary files
        for temp_file in ["chapter_list.json", "cover.jpg"]:
            if os.path.exists(temp_file):
                try:
                    os.remove(temp_file)
                except Exception:
                    pass
        logger.info("Đã dọn dẹp file tạm. Hẹn gặp lại!")


if __name__ == "__main__":
    main()



import asyncio
import httpx
from playwright.async_api import async_playwright # Added for dynamic content
from bs4 import BeautifulSoup
import json
import subprocess # Added for notify-send
from typing import Optional, Dict, Any

from .utils import HEADERS


async def get_chapter_tree(url: str, output_file: str, cookies: Optional[Dict[str, str]] = None):
    print("Đang tạo sơ đồ cây...")
    """
    Sử dụng httpx để truy cập URL, sau đó dùng BeautifulSoup để
    phân tích và trích xuất sơ đồ các tập và chương truyện, rồi lưu vào file txt.
    Phiên bản này tương thích với môi trường đã có asyncio loop.

    Args:
        url (str): URL của trang truyện.
        output_file (str): Tên của file txt để lưu sơ đồ.
    """
    try:
        async with httpx.AsyncClient(headers=HEADERS, cookies=cookies) as client:
            response = await client.get(url, follow_redirects=True)
            response.raise_for_status() # Raise an exception for 4xx or 5xx status codes
            html_content = response.text

        soup = BeautifulSoup(html_content, 'html.parser')
        
        chapter_tree_string = ""
        volumes = soup.find_all('div', class_='module-container')

        if not volumes:
            print("Không tìm thấy container nào cho các tập truyện.")
            return

        print(f"Tìm thấy {len(volumes)} tập/phần truyện. Bắt đầu trích xuất...")

        for volume in volumes:
            volume_title_element = volume.find('h3', class_='module-title')
            if volume_title_element:
                volume_title = volume_title_element.get_text(strip=True)
                chapter_tree_string += f"■ {volume_title}\n"
            else:
                chapter_tree_string += "■ [Không có tiêu đề tập]\n"

            chapters = volume.find_all('div', class_='module-chapter-item')
            if chapters:
                for chapter in chapters:
                    chapter_link = chapter.find('a', class_='chapter-title-link')
                    if chapter_link:
                        chapter_title = chapter_link.get_text(strip=True)
                        chapter_tree_string += f"  - {chapter_title}\n"
            else:
                 chapter_tree_string += "  - [Không có chương nào trong tập này]\n"
            
            chapter_tree_string += "\n"

        with open(output_file, "w", encoding="utf-8") as f:
            f.write(chapter_tree_string)
        
        print(f"Đã tạo thành công sơ đồ các chương và lưu vào file '{output_file}'")

    except Exception as e:
        print(f"Đã xảy ra lỗi: {e}")

async def get_chapter_tree_folder(url: str, output_file: str, cookies: Optional[Dict[str, str]] = None):
    print("Đang tạo thư mục...")
    """
    Sử dụng httpx để truy cập URL, sau đó dùng BeautifulSoup để
    phân tích và trích xuất sơ đồ các tập và chương truyện, rồi lưu vào file txt.
    Phiên bản này tương thích với môi trường đã có asyncio loop.

    Args:
        url (str): URL của trang truyện.
        output_file (str): Tên của file txt để lưu sơ đồ.
    """
    try:
        async with httpx.AsyncClient(headers=HEADERS, cookies=cookies) as client:
            response = await client.get(url, follow_redirects=True)
            response.raise_for_status()
            html_content = response.text

        soup = BeautifulSoup(html_content, 'html.parser')
        
        chapter_tree_string = ""
        volumes = soup.find_all('div', class_='module-container')

        if not volumes:
            print("Không tìm thấy container nào cho các tập truyện.")
            return

        print(f"Tìm thấy {len(volumes)} tập/phần truyện. Bắt đầu trích xuất...")

        for volume in volumes:
            volume_title_element = volume.find('h3', class_='module-title')
            if volume_title_element:
                volume_title = volume_title_element.get_text(strip=True)
                volume_title_string = volume_title.replace(":", " -").replace("/", " -").replace("\\", " -").replace("*", " -").replace("?", " -").replace("\"", " -").replace("<", " -").replace(">", " -").replace("|", " -")
                chapter_tree_string += f" {volume_title_string}\n"
            else:
                chapter_tree_string += "[no name]\n"
            chapter_tree_string += "\n"

        with open(output_file, "w", encoding="utf-8") as f:
            f.write(chapter_tree_string)
        
        #print(f"Đã tạo thành công sơ đồ các chương và lưu vào file '{output_file}'")

    except Exception as e:
        print(f"Đã xảy ra lỗi: {e}")
#creat list chapter
async def get_chapter_tree_list(url: str, output_file: str = "chapter_list.json", session_state: Optional[Dict[str, Any]] = None):
    print("Đang tạo sơ đồ cây (sử dụng Playwright cho nội dung động)...")

    try:
        async with async_playwright() as p:
            browser = await p.chromium.launch()
            # Use consistent User-Agent
            user_agent = HEADERS.get("User-Agent")
            context = await browser.new_context(storage_state=session_state, user_agent=user_agent) if session_state else await browser.new_context(user_agent=user_agent)
            page = await context.new_page()
            await page.goto(url, wait_until='domcontentloaded', timeout=60000)
            
            # Wait for content to stabilize
            await page.wait_for_timeout(2000)
            await page.wait_for_selector('.module-chapter-item', timeout=30000)
            
            html_content = await page.content()
            await browser.close()

        soup = BeautifulSoup(html_content, 'html.parser')
        volumes = soup.find_all('div', class_='module-container')

        if not volumes:
            print("Không tìm thấy container nào cho các tập truyện.")
            return []

        print(f"Tìm thấy {len(volumes)} tập/phần truyện. Bắt đầu trích xuất...")

        data = []

        for volume in volumes:
            volume_title_element = volume.find('h3', class_='module-title')
            if volume_title_element:
                volume_title = volume_title_element.get_text(strip=True)
            else:
                volume_title = "[Không có tiêu đề tập]"

            chapters_list = []
            chapters = volume.find_all('div', class_='module-chapter-item')
            if chapters:
                for chapter in chapters:
                    try:
                        # Find link - try both specific class and any anchor tag
                        link_element = chapter.find('a', class_='chapter-title-link') or chapter.find('a')

                        if link_element and 'href' in link_element.attrs:
                            chapter_link = link_element['href']
                            chapter_title = link_element.get_text(strip=True)
                            if "minh-hoa" not in chapter_link:
                                chapters_list.append({
                                    "title": chapter_title,
                                    "url": chapter_link
                                })
                        else:
                            # If we still can't find it, log the classes for debug
                            classes = chapter.get('class', [])
                            if 'locked-chapter' in classes or 'chapter-mode-protected' in classes:
                                print(f"[Cảnh báo] Chương này vẫn bị khóa (locked): {chapter.get_text(strip=True)}")
                            else:
                                print(f"Cant find link for chapter: {chapter}")
                    except Exception:
                        pass # Silently skip malformed chapters

            data.append({
                "volume": volume_title,
                "chapters": chapters_list
            })

        # Lưu ra file JSON
        with open(output_file, "w", encoding="utf-8") as f:
            json.dump(data, f, ensure_ascii=False, indent=2)
        print(f"Đã lưu sơ đồ cây vào {output_file}")
        return data

    except Exception as e:
        print(f"Đã xảy ra lỗi khi dùng Playwright để lấy danh sách chương: {e}")
        return []

def get_chapters_by_volume_index(file_path: str, index: int):
    try:
        with open(file_path, "r", encoding="utf-8") as f:
            data = json.load(f)

        if index < 0 or index >= len(data):
            print("Index không hợp lệ. Trong file chỉ có", len(data), "tập.")
            return []

        volume = data[index]
        #print(f"Tên tập: {volume['volume']}")
        return volume["chapters"]

    except Exception as e:
        print("Đã xảy ra lỗi khi đọc file:", e)
        return []



"""
File export functions for the web novel scraper.
"""
import os
import asyncio
import urllib.parse
from io import BytesIO
from typing import List, Dict, Any, Union, cast, Optional

import httpx
# Heavy AI libraries (numpy, vieneu) are lazy-loaded inside tao_file_mp3 
# to ensure a fast cold start for the CLI and Web UI.
from ebooklib import epub
from reportlab.platypus import SimpleDocTemplate, Paragraph, Spacer, Image
from reportlab.lib.styles import getSampleStyleSheet, ParagraphStyle
from reportlab.lib.units import inch
from reportlab.pdfbase import pdfmetrics
from reportlab.pdfbase.ttfonts import TTFont
from PIL import Image as PILImage
from loguru import logger
from rich.progress import Progress, SpinnerColumn, TextColumn, BarColumn, TaskProgressColumn

from .utils import HEADERS
from .models import StoryInfo, ContentItem, Chapter, Volume


ContentItemLike = Union[ContentItem, Dict[str, str]]
ContentList = List[ContentItemLike]
ChapterData = Dict[str, Any]  # {'title': str, 'content': ContentList}
VolumeData = Dict[str, Any]   # {'volume': str, 'chapters': List[ChapterData]}
ChaptersData = List[Union[ChapterData, VolumeData]]


def _normalize_content_item(item: ContentItemLike) -> ContentItem:
    """Convert dict to ContentItem if needed."""
    if isinstance(item, ContentItem):
        return item
    return ContentItem(type=cast(str, item['type']), data=cast(str, item['data']))


def _normalize_content_list(items: ContentList) -> List[ContentItem]:
    """Convert list of dicts to list of ContentItems."""
    return [_normalize_content_item(item) for item in items]


async def _download_images_bulk(urls: List[str], max_concurrent: int = 10) -> Dict[str, bytes]:
    """Downloads multiple images concurrently with a limit on parallelism and a progress bar."""
    if not urls:
        return {}

    unique_urls = list(set(urls))
    semaphore = asyncio.Semaphore(max_concurrent)
    results = {}

    async with httpx.AsyncClient(headers=HEADERS, timeout=30.0, follow_redirects=True) as client:
        with Progress(
            SpinnerColumn(),
            TextColumn("[progress.description]{task.description}"),
            BarColumn(),
            TaskProgressColumn(),
            transient=True
        ) as progress:
            task = progress.add_task("[cyan]Đang tải ảnh minh họa...", total=len(unique_urls))
            
            async def download(url: str):
                async with semaphore:
                    try:
                        response = await client.get(url)
                        response.raise_for_status()
                        progress.update(task, advance=1)
                        return url, response.content
                    except Exception as e:
                        logger.warning(f"Lỗi khi tải ảnh {url}: {e}")
                        progress.update(task, advance=1)
                        return url, None

            tasks = [download(url) for url in unique_urls]
            downloaded = await asyncio.gather(*tasks)
            for url, content in downloaded:
                if content:
                    results[url] = content
    return results


def _get_image_extension(url: str, content_type: Optional[str] = None) -> str:
    """Detects image extension from URL or Content-Type header."""
    if content_type:
        if 'jpeg' in content_type: return 'jpg'
        if 'png' in content_type: return 'png'
        if 'gif' in content_type: return 'gif'
        if 'svg' in content_type: return 'svg'
    
    parsed_url = urllib.parse.urlparse(url)
    ext = parsed_url.path.split('.')[-1].lower()
    return ext if ext in ['jpg', 'jpeg', 'png', 'gif', 'svg'] else 'jpg'


async def tao_file_epub(
    filename: str,
    book_title: str,
    author: str,
    chapters_data: ChaptersData,
    description: str = "",
    cover_path: Union[str, None] = None,
    genres: List[str] = None
) -> None:
    """Creates an EPUB file with pre-fetched images for high performance."""
    logger.info(f"Đang tạo file EPUB: {filename}")
    
    # 1. Extract all image URLs
    all_image_urls = []
    for item in chapters_data:
        chaps = item.get('chapters', []) if 'chapters' in item else [item]
        for chap in chaps:
            for ci in chap.get('content', []):
                norm = _normalize_content_item(ci)
                if norm.type == 'image': all_image_urls.append(norm.data)
    
    # 2. Pre-download all images concurrently
    image_cache = await _download_images_bulk(all_image_urls)

    # 3. Build EPUB
    book = epub.EpubBook()
    book.set_identifier(f'urn:uuid:{os.path.basename(filename)}')
    book.set_title(book_title)
    book.set_language('vi')
    book.add_author(author)
    book.add_metadata('DC', 'description', description)
    if genres:
        for g in genres: book.add_metadata('DC', 'subject', g)

    try:
        book.set_cover("cover.jpg", open('cover.jpg', 'rb').read())
    except: pass

    toc = []
    spine = ['nav']
    url_to_internal_path = {}
    image_counter = 1

    # Map pre-downloaded images to EPUB items
    for url, content in image_cache.items():
        ext = _get_image_extension(url)
        img_name = f'image_{image_counter}.{ext}'
        img_item = epub.EpubImage(
            uid=f'img_{image_counter}',
            file_name=f'images/{img_name}',
            media_type=f'image/{ext}',
            content=content
        )
        book.add_item(img_item)
        url_to_internal_path[url] = f'images/{img_name}'
        image_counter += 1

    def process_chapter(chap_data: ChapterData, chap_idx: int) -> epub.EpubHtml:
        title = chap_data.get('title', f"Chương {chap_idx}")
        chapter_obj = epub.EpubHtml(title=title, file_name=f'chap_{chap_idx}.xhtml', lang='vi')
        html = f'<h1>{title}</h1>'
        for item in chap_data.get('content', []):
            norm = _normalize_content_item(item)
            if norm.type == 'text':
                html += f'<p>{norm.data}</p>'
            elif norm.type == 'image' and norm.data in url_to_internal_path:
                html += f'<img src="{url_to_internal_path[norm.data]}" alt="Minh họa"/>'
        chapter_obj.content = html
        return chapter_obj

    # Assemble structure
    chapter_index = 1
    for item in chapters_data:
        if 'volume' in item:
            vol_title = item['volume']
            vol_chaps = []
            for c_data in item.get('chapters', []):
                chap_obj = process_chapter(c_data, chapter_index)
                book.add_item(chap_obj)
                spine.append(chap_obj)
                vol_chaps.append(epub.Link(chap_obj.file_name, chap_obj.title, f'ch_{chapter_index}'))
                chapter_index += 1
            toc.append((epub.Section(vol_title), tuple(vol_chaps)))
        else:
            chap_obj = process_chapter(item, chapter_index)
            book.add_item(chap_obj)
            spine.append(chap_obj)
            toc.append(epub.Link(chap_obj.file_name, chap_obj.title, f'ch_{chapter_index}'))
            chapter_index += 1

    book.toc = tuple(toc)
    book.spine = spine
    book.add_item(epub.EpubNcx())
    book.add_item(epub.EpubNav())
    epub.write_epub(filename, book, {})
    logger.info(f"Tạo file EPUB thành công: {filename}")


async def tao_file_pdf(
    content_list: ContentList,
    filename: str,
    title: str = "Chương truyện",
    font_name: str = 'DejaVuSans'
) -> None:
    """Creates a PDF with optimized image fetching."""
    logger.info(f"Đang tạo file PDF: {filename}")
    
    # 1. Prepare Font
    font_path = f"{font_name}.ttf"
    if not os.path.exists(font_path):
        font_urls = {
            'DejaVuSans': 'https://github.com/dejavu-fonts/dejavu-fonts/raw/master/ttf/DejaVuSans.ttf',
            'NotoSerif': 'https://raw.githubusercontent.com/google/fonts/main/ofl/notoserif/NotoSerif-Regular.ttf'
        }
        if font_name in font_urls:
            async with httpx.AsyncClient() as client:
                resp = await client.get(font_urls[font_name])
                with open(font_path, 'wb') as f: f.write(resp.content)

    # 2. Pre-download images
    normalized_items = _normalize_content_list(content_list)
    img_urls = [i.data for i in normalized_items if i.type == 'image']
    image_cache = await _download_images_bulk(img_urls)

    # 3. Build PDF
    try:
        pdfmetrics.registerFont(TTFont(font_name, font_path))
        style = ParagraphStyle(name='Normal_vi', fontName=font_name, fontSize=12, leading=14)
        title_style = ParagraphStyle(name='Title_vi', fontName=font_name, fontSize=18, leading=22, spaceAfter=0.2 * inch)
    except:
        styles = getSampleStyleSheet()
        style, title_style = styles['Normal'], styles['h1']

    doc = SimpleDocTemplate(filename)
    story = [Paragraph(title, title_style), Spacer(1, 0.2 * inch)]
    max_w, max_h = doc.width, doc.height

    for item in normalized_items:
        if item.type == 'text':
            story.append(Paragraph(item.data, style))
            story.append(Spacer(1, 0.1 * inch))
        elif item.type == 'image' and item.data in image_cache:
            try:
                img_data = BytesIO(image_cache[item.data])
                pil_img = PILImage.open(img_data)
                w, h = pil_img.size
                ratio = min(max_w/w, max_h/h, 1)
                story.append(Image(img_data, width=w*ratio, height=h*ratio))
                story.append(Spacer(1, 0.1 * inch))
            except: pass

    doc.build(story)
    logger.info(f"Tạo file PDF thành công: {filename}")


async def tao_file_html(content_list: ContentList, filename: str, title: str = "Chương truyện") -> None:
    """HTML export (Async for API consistency)."""
    logger.info(f"Đang tạo file HTML: {filename}")
    html = f"<!DOCTYPE html><html lang='vi'><head><meta charset='UTF-8'><title>{title}</title></head><body><h1>{title}</h1>"
    for item in _normalize_content_list(content_list):
        if item.type == 'text': html += f"<p>{item.data}</p>"
        elif item.type == 'image': html += f"<img src='{item.data}' style='max-width:100%'/>"
    html += "</body></html>"
    with open(filename, 'w', encoding='utf-8') as f: f.write(html)
    logger.info(f"Tạo file HTML thành công: {filename}")

async def tao_file_md(content_list: ContentList, filename: str, title: str = "Chương truyện") -> None:
    """Markdown export (Async for API consistency)."""
    logger.info(f"Đang tạo file Markdown: {filename}")
    md = f"# {title}\n\n"
    for item in _normalize_content_list(content_list):
        if item.type == 'text': md += f"{item.data}\n\n"
        elif item.type == 'image': md += f"![Minh họa]({item.data})\n\n"
    with open(filename, 'w', encoding='utf-8') as f: f.write(md)
    logger.info(f"Tạo file Markdown thành công: {filename}")

async def tao_file_txt(content_list: ContentList, filename: str, title: str = "Chương truyện") -> None:
    """Text export (Async for API consistency)."""
    logger.info(f"Đang tạo file Text: {filename}")
    txt = f"{title}\n\n"
    for item in _normalize_content_list(content_list):
        if item.type == 'text': txt += f"{item.data}\n\n"
        elif item.type == 'image': txt += f"[Ảnh: {item.data}]\n\n"
    with open(filename, 'w', encoding='utf-8') as f: f.write(txt)
    logger.info(f"Tạo file Text thành công: {filename}")


async def tao_file_mp3(content_list: ContentList, filename: str, title: str = "Chương truyện") -> None:
    """AI-Powered Audiobook generation using VieNeu with chunked processing."""
    # 1. SILENCE HEAVY WARNINGS BEFORE IMPORTING
    import os
    import warnings
    # Suppress heavy framework logs and warnings to keep the terminal clean,
    # but let PyTorch/Tensorflow decide hardware usage (CPU vs GPU) automatically.
    os.environ["TF_CPP_MIN_LOG_LEVEL"] = "3"
    warnings.filterwarnings("ignore", category=UserWarning)

    # 2. LAZY LOAD HEAVY LIBRARIES
    try:
        import numpy as np
        from vieneu import Vieneu
    except ImportError:
        logger.error("Vieneu or numpy not found. Please run 'pip install vieneu numpy' to use TTS.")
        return

    logger.info(f"Đang tạo file Audiobook: {filename} (Sử dụng VieNeu AI)")
    
    # 3. Prepare text chunks
    chunks = [title]
    for item in _normalize_content_list(content_list):
        if item.type == 'text':
            text = item.data.strip()
            if text:
                if len(text) > 2000:
                    subchunks = [text[i:i+2000] for i in range(0, len(text), 2000)]
                    chunks.extend(subchunks)
                else:
                    chunks.append(text)

    # 4. Synthesize each chunk
    try:
        def run_tts_chunked():
            tts = Vieneu()
            voice_data = tts.get_preset_voice("Tuyen")
            audio_segments = []
            
            total_chunks = len(chunks)
            for i, chunk in enumerate(chunks):
                if not chunk.strip():
                    continue
                logger.debug(f"Synthesizing chunk {i+1}/{total_chunks}...")
                audio = tts.infer(text=chunk, voice=voice_data)
                audio_segments.append(audio)
            
            if audio_segments:
                logger.debug("Merging audio segments...")
                merged_audio = np.concatenate(audio_segments)
                tts.save(merged_audio, filename)
            
        await asyncio.to_thread(run_tts_chunked)
        logger.info(f"Tạo file Audiobook thành công: {filename}")
    except Exception as e:
        logger.error(f"Lỗi khi tạo Audiobook: {e}")
        raise e



"""
Data models for the web novel scraper.
"""
from dataclasses import dataclass
from typing import Dict, List, Optional, Union, Literal, TypedDict, Any


@dataclass
class StoryInfo:
    """Information about a story."""
    title: str
    author: str
    description: str
    genres: List[str] = None
    cover_path: Optional[str] = None


@dataclass
class ContentItem:
    """A single content item (text or image)."""
    type: Literal["text", "image"]
    data: str


@dataclass
class Chapter:
    """A chapter with title and content."""
    title: str
    content: List[ContentItem]
    url: Optional[str] = None


@dataclass
class Volume:
    """A volume containing chapters."""
    title: str
    chapters: List[Chapter]


# Type aliases for backward compatibility
ChapterData = Dict[str, Union[str, List[Dict[str, str]]]]
VolumeData = Dict[str, Union[str, List[ChapterData]]]
StoryInfoDict = Dict[str, Any]


def story_info_to_dict(info: StoryInfo) -> StoryInfoDict:
    """Convert StoryInfo to dictionary for backward compatibility."""
    return {
        "title": info.title,
        "author": info.author,
        "description": info.description,
        "genres": info.genres or [],
        "cover_path": info.cover_path
    }


def dict_to_story_info(data: StoryInfoDict) -> StoryInfo:
    """Convert dictionary to StoryInfo."""
    return StoryInfo(
        title=data.get("title") or "Unknown Title",
        author=data.get("author") or "Unknown Author",
        description=data.get("description") or "No Description",
        genres=data.get("genres") or [],
        cover_path=data.get("cover_path")
    )


import os
import json
import pytest
from vvr_scraper.session_manager import save_session, load_session

def test_save_load_session(tmp_path):
    session_file = tmp_path / ".vvr_session.json"
    test_state = {"cookies": [{"name": "token", "value": "test_token"}], "origins": []}
    save_session(test_state, str(session_file))
    assert session_file.exists()
    loaded_state = load_session(str(session_file))
    assert loaded_state == test_state



"""
Tests for tao_so_do_cay.py - Chapter tree extraction utilities
"""
import pytest
import os
import sys
import json
import tempfile
from unittest.mock import AsyncMock, MagicMock, patch

# Add parent directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))
from vvr_scraper.tao_so_do_cay import (
    get_chapter_tree,
    get_chapter_tree_folder,
    get_chapter_tree_list,
    get_chapters_by_volume_index
)
from vvr_scraper.utils import HEADERS


# =============================================================================
# TESTS FOR get_chapter_tree
# =============================================================================

class TestGetChapterTree:
    """Tests for the get_chapter_tree function"""

    @pytest.mark.asyncio
    async def test_basic_chapter_extraction(self, tmp_path):
        """Test basic chapter tree extraction from HTML"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1</h3>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-1">Chương 1</a>
                </div>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-2">Chương 2</a>
                </div>
            </div>
        </html>
        """

        output_file = str(tmp_path / "tree.txt")

        with patch('httpx.AsyncClient') as MockClient:
            mock_response = MagicMock()
            mock_response.text = mock_html
            mock_response.raise_for_status = MagicMock()

            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            await get_chapter_tree("https://example.com/story", output_file)

            assert os.path.exists(output_file)
            with open(output_file, 'r', encoding='utf-8') as f:
                content = f.read()

            assert 'Volume 1' in content
            assert 'Chương 1' in content
            assert 'Chương 2' in content

    @pytest.mark.asyncio
    async def test_no_volumes_found(self, capsys):
        """Test handling when no volumes are found"""
        mock_html = "<html><body>No content</body></html>"

        with patch('httpx.AsyncClient') as MockClient:
            mock_response = MagicMock()
            mock_response.text = mock_html
            mock_response.raise_for_status = MagicMock()

            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            await get_chapter_tree("https://example.com/story", "/tmp/test.txt")

            captured = capsys.readouterr()
            assert "Không tìm thấy container nào" in captured.out

    @pytest.mark.asyncio
    async def test_volume_without_title(self, tmp_path):
        """Test handling volumes without titles"""
        mock_html = """
        <html>
            <div class="module-container">
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-1">Chương 1</a>
                </div>
            </div>
        </html>
        """

        output_file = str(tmp_path / "tree.txt")

        with patch('httpx.AsyncClient') as MockClient:
            mock_response = MagicMock()
            mock_response.text = mock_html
            mock_response.raise_for_status = MagicMock()

            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            await get_chapter_tree("https://example.com/story", output_file)

            with open(output_file, 'r', encoding='utf-8') as f:
                content = f.read()

            assert '[Không có tiêu đề tập]' in content

    @pytest.mark.asyncio
    async def test_empty_volume(self, tmp_path):
        """Test handling volumes with no chapters"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Empty Volume</h3>
            </div>
        </html>
        """

        output_file = str(tmp_path / "tree.txt")

        with patch('httpx.AsyncClient') as MockClient:
            mock_response = MagicMock()
            mock_response.text = mock_html
            mock_response.raise_for_status = MagicMock()

            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            await get_chapter_tree("https://example.com/story", output_file)

            with open(output_file, 'r', encoding='utf-8') as f:
                content = f.read()

            assert '[Không có chương nào trong tập này]' in content


# =============================================================================
# TESTS FOR get_chapter_tree_folder
# =============================================================================

class TestGetChapterTreeFolder:
    """Tests for the get_chapter_tree_folder function"""

    @pytest.mark.asyncio
    async def test_sanitizes_volume_titles(self, tmp_path):
        """Test that volume titles are sanitized for folder names"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1: Special*Chars?</h3>
            </div>
        </html>
        """

        output_file = str(tmp_path / "tree.txt")

        with patch('httpx.AsyncClient') as MockClient:
            mock_response = MagicMock()
            mock_response.text = mock_html
            mock_response.raise_for_status = MagicMock()

            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            await get_chapter_tree_folder("https://example.com/story", output_file)

            with open(output_file, 'r', encoding='utf-8') as f:
                content = f.read()

            # Special characters should be replaced with " -"
            assert '*' not in content
            assert '?' not in content
            assert ':' not in content

    @pytest.mark.asyncio
    async def test_multiple_volumes(self, tmp_path):
        """Test extraction of multiple volumes"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1</h3>
            </div>
            <div class="module-container">
                <h3 class="module-title">Volume 2</h3>
            </div>
            <div class="module-container">
                <h3 class="module-title">Volume 3</h3>
            </div>
        </html>
        """

        output_file = str(tmp_path / "tree.txt")

        with patch('httpx.AsyncClient') as MockClient:
            mock_response = MagicMock()
            mock_response.text = mock_html
            mock_response.raise_for_status = MagicMock()

            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            await get_chapter_tree_folder("https://example.com/story", output_file)

            with open(output_file, 'r', encoding='utf-8') as f:
                content = f.read()

            assert 'Volume 1' in content
            assert 'Volume 2' in content
            assert 'Volume 3' in content


# =============================================================================
# TESTS FOR get_chapter_tree_list
# =============================================================================

class TestGetChapterTreeList:
    """Tests for the get_chapter_tree_list function (uses Playwright)"""

    def _setup_mock_playwright(self, mock_playwright, mock_html):
        mock_browser = AsyncMock()
        mock_context = AsyncMock()
        mock_page = AsyncMock()
        mock_page.content = AsyncMock(return_value=mock_html)

        mock_p_instance = MagicMock()
        mock_p_instance.chromium.launch = AsyncMock(return_value=mock_browser)

        mock_browser.__aenter__ = AsyncMock(return_value=mock_browser)
        mock_browser.__aexit__ = AsyncMock(return_value=None)
        mock_browser.new_context = AsyncMock(return_value=mock_context)
        mock_context.new_page = AsyncMock(return_value=mock_page)

        mock_playwright.return_value.__aenter__.return_value = mock_p_instance
        mock_playwright.return_value.__aexit__ = AsyncMock(return_value=None)
        return mock_browser, mock_context, mock_page

    @pytest.mark.asyncio
    async def test_creates_json_output(self, tmp_path):
        """Test that JSON output file is created with correct structure"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1</h3>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-1">Chương 1</a>
                </div>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-2">Chương 2</a>
                </div>
            </div>
        </html>
        """

        output_file = str(tmp_path / "chapters.json")

        with patch('vvr_scraper.tao_so_do_cay.async_playwright') as mock_playwright:
            self._setup_mock_playwright(mock_playwright, mock_html)
            result = await get_chapter_tree_list("https://example.com/story", output_file)

            assert os.path.exists(output_file)
            with open(output_file, 'r', encoding='utf-8') as f:
                data = json.load(f)

            assert len(data) == 1
            assert data[0]['volume'] == 'Volume 1'
            assert len(data[0]['chapters']) == 2
            assert data[0]['chapters'][0]['title'] == 'Chương 1'
            assert data[0]['chapters'][0]['url'] == '/chap-1'

    @pytest.mark.asyncio
    async def test_crawls_both_published_and_protected_chapters(self, tmp_path):
        """Test that both published and protected chapters are crawled"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1</h3>
                <div class="module-chapter-item chapter-item-animated chapter-mode-published">
                    <a class="chapter-title-link" href="/chap-published">Published</a>
                </div>
                <div class="module-chapter-item chapter-item-animated chapter-mode-protected">
                    <a class="chapter-title-link" href="/chap-protected">Protected</a>
                </div>
            </div>
        </html>
        """

        output_file = str(tmp_path / "chapters_mixed.json")

        with patch('vvr_scraper.tao_so_do_cay.async_playwright') as mock_playwright:
            self._setup_mock_playwright(mock_playwright, mock_html)
            result = await get_chapter_tree_list("https://example.com/story", output_file)

            assert len(result[0]['chapters']) == 2
            urls = [c['url'] for c in result[0]['chapters']]
            assert '/chap-published' in urls
            assert '/chap-protected' in urls

    @pytest.mark.asyncio
    async def test_uses_session_state(self, tmp_path):
        """Test that session_state is passed to Playwright context"""
        mock_html = "<html><div class='module-chapter-item'><a class='chapter-title-link' href='/c1'>C1</a></div></html>"
        output_file = str(tmp_path / "session_test.json")
        session_state = {"cookies": [{"name": "test", "value": "val"}]}

        with patch('vvr_scraper.tao_so_do_cay.async_playwright') as mock_playwright:
            mock_browser, _, _ = self._setup_mock_playwright(mock_playwright, mock_html)
            await get_chapter_tree_list("https://example.com/story", output_file, session_state=session_state)
            mock_browser.new_context.assert_called_with(storage_state=session_state, user_agent=HEADERS.get("User-Agent"))

    @pytest.mark.asyncio
    async def test_filters_minh_hoa_chapters(self, tmp_path):
        """Test that 'minh-hoa' (illustration) chapters are filtered out"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1</h3>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-1">Chương 1</a>
                </div>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-2-minh-hoa">Minh Họa</a>
                </div>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/chap-3">Chương 3</a>
                </div>
            </div>
        </html>
        """

        output_file = str(tmp_path / "chapters_filter.json")

        with patch('vvr_scraper.tao_so_do_cay.async_playwright') as mock_playwright:
            self._setup_mock_playwright(mock_playwright, mock_html)
            result = await get_chapter_tree_list("https://example.com/story", output_file)

            assert len(result[0]['chapters']) == 2
            urls = [c['url'] for c in result[0]['chapters']]
            assert '/chap-2-minh-hoa' not in urls

    @pytest.mark.asyncio
    async def test_handles_missing_href(self, tmp_path):
        """Test handling of chapters without href attribute"""
        mock_html = """
        <html>
            <div class="module-container">
                <h3 class="module-title">Volume 1</h3>
                <div class="module-chapter-item">
                    <a class="chapter-title-link">No href</a>
                </div>
                <div class="module-chapter-item">
                    <a class="chapter-title-link" href="/valid-chap">Valid</a>
                </div>
            </div>
        </html>
        """

        output_file = str(tmp_path / "chapters_href.json")

        with patch('vvr_scraper.tao_so_do_cay.async_playwright') as mock_playwright:
            self._setup_mock_playwright(mock_playwright, mock_html)
            result = await get_chapter_tree_list("https://example.com/story", output_file)

            assert len(result[0]['chapters']) == 1
            assert result[0]['chapters'][0]['url'] == '/valid-chap'


# =============================================================================
# TESTS FOR get_chapters_by_volume_index
# =============================================================================

class TestGetChaptersByVolumeIndex:
    """Tests for the get_chapters_by_volume_index function"""

    def test_valid_index(self, tmp_path):
        """Test getting chapters with valid index"""
        test_data = [
            {"volume": "Volume 1", "chapters": [{"title": "C1", "url": "/chap-1"}, {"title": "C2", "url": "/chap-2"}]},
            {"volume": "Volume 2", "chapters": [{"title": "C3", "url": "/chap-3"}, {"title": "C4", "url": "/chap-4"}]},
        ]

        json_file = str(tmp_path / "test.json")
        with open(json_file, 'w', encoding='utf-8') as f:
            json.dump(test_data, f, ensure_ascii=False)

        result = get_chapters_by_volume_index(json_file, 0)
        assert result == [{"title": "C1", "url": "/chap-1"}, {"title": "C2", "url": "/chap-2"}]

    def test_second_volume(self, tmp_path):
        """Test getting chapters from second volume"""
        test_data = [
            {"volume": "Volume 1", "chapters": [{"title": "C1", "url": "/chap-1"}, {"title": "C2", "url": "/chap-2"}]},
            {"volume": "Volume 2", "chapters": [{"title": "C3", "url": "/chap-3"}, {"title": "C4", "url": "/chap-4"}]},
        ]

        json_file = str(tmp_path / "test.json")
        with open(json_file, 'w', encoding='utf-8') as f:
            json.dump(test_data, f, ensure_ascii=False)

        result = get_chapters_by_volume_index(json_file, 1)
        assert result == [{"title": "C3", "url": "/chap-3"}, {"title": "C4", "url": "/chap-4"}]

    def test_invalid_index_negative(self, capsys):
        """Test handling of negative index"""
        test_data = [{"volume": "Volume 1", "chapters": ["/chap-1"]}]

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(test_data, f)
            json_file = f.name

        try:
            result = get_chapters_by_volume_index(json_file, -1)
            assert result == []
            captured = capsys.readouterr()
            assert "không hợp lệ" in captured.out
        finally:
            os.unlink(json_file)

    def test_invalid_index_out_of_range(self, capsys):
        """Test handling of index out of range"""
        test_data = [{"volume": "Volume 1", "chapters": ["/chap-1"]}]

        with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as f:
            json.dump(test_data, f)
            json_file = f.name

        try:
            result = get_chapters_by_volume_index(json_file, 100)
            assert result == []
            captured = capsys.readouterr()
            assert "không hợp lệ" in captured.out
        finally:
            os.unlink(json_file)

    def test_nonexistent_file(self, capsys):
        """Test handling of nonexistent file"""
        result = get_chapters_by_volume_index("/nonexistent/file.json", 0)
        assert result == []
        captured = capsys.readouterr()
        assert "Đã xảy ra lỗi" in captured.out


# =============================================================================
# TESTS FOR HEADERS CONFIGURATION
# =============================================================================

class TestTaoSoDoCayHeaders:
    """Tests for HEADERS configuration in tao_so_do_cay"""

    def test_headers_contains_required_fields(self):
        """Test that HEADERS contains all required fields"""
        assert 'User-Agent' in HEADERS
        assert 'Accept' in HEADERS
        assert 'Accept-Language' in HEADERS

    def test_user_agent_is_browser_like(self):
        """Test that User-Agent looks like a real browser"""
        ua = HEADERS['User-Agent']
        assert 'Mozilla' in ua
        assert 'Chrome' in ua



"""
FastAPI web server for the Valvrare Team Scraper.
"""
import asyncio
import os
import uuid
from typing import List, Optional, Dict, Any
from contextlib import asynccontextmanager

import httpx
import uvicorn
from fastapi import FastAPI, WebSocket, WebSocketDisconnect, Query
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
from pydantic import BaseModel
from loguru import logger
import json

from .scraper_core import lay_thong_tin_truyen, scrape_chapters
from .exporter import (
    tao_file_epub, tao_file_pdf, tao_file_html, tao_file_md, tao_file_txt, tao_file_mp3
)
from .utils import (
    sanitize_filename, HEADERS, normalize_vietnamese_url, get_token_from_state
)
from .session_manager import load_session, save_session
from .tao_so_do_cay import get_chapter_tree_list
from playwright.async_api import async_playwright

class DownloadManager:
    def __init__(self, num_workers: int = 1):
        self.queue = asyncio.Queue()
        self.workers = []
        self.num_workers = num_workers

    async def start_workers(self):
        logger.info(f"Starting {self.num_workers} download workers...")
        for _ in range(self.num_workers):
            worker = asyncio.create_task(self.worker_loop())
            self.workers.append(worker)

    async def stop_workers(self):
        logger.info("Stopping download workers...")
        for worker in self.workers:
            worker.cancel()
        await asyncio.gather(*self.workers, return_exceptions=True)

    async def worker_loop(self):
        while True:
            try:
                req, task_id = await self.queue.get()
                # Use contextualize to bind task_id to logs in this worker
                with logger.contextualize(task_id=task_id):
                    await run_scrape_task(req, task_id)
            except asyncio.CancelledError:
                break
            except Exception as e:
                logger.error(f"Worker encountered error: {e}")
            finally:
                self.queue.task_done()

    async def add_task(self, req: "DownloadRequest", task_id: str):
        await manager.broadcast({"type": "status", "task_id": task_id, "status": "In Queue..."})
        await self.queue.put((req, task_id))

download_queue = DownloadManager(num_workers=1)

@asynccontextmanager
async def lifespan(app: FastAPI):
    # Startup
    await download_queue.start_workers()
    yield
    # Shutdown
    await download_queue.stop_workers()

app = FastAPI(title="Valvrare Team Scraper Web UI", lifespan=lifespan)

BASE_URL = "https://valvrareteam.net"
SSR_API_URL = "https://val-ssr-2kzit.ondigitalocean.app/api/novels/search"

class ConnectionManager:
    def __init__(self):
        self.active_connections: List[WebSocket] = []

    async def connect(self, websocket: WebSocket):
        await websocket.accept()
        self.active_connections.append(websocket)

    def disconnect(self, websocket: WebSocket):
        if websocket in self.active_connections:
            self.active_connections.remove(websocket)

    async def broadcast(self, message: dict):
        for connection in self.active_connections:
            try:
                await connection.send_json(message)
            except Exception:
                pass

manager = ConnectionManager()

def websocket_sink(message):
    """Loguru sink that broadcasts logs via WebSocket."""
    record = message.record
    task_id = record["extra"].get("task_id", "system")
    log_msg = {
        "type": "log",
        "task_id": task_id,
        "level": record["level"].name,
        "message": record["message"],
        "time": record["time"].strftime("%H:%M:%S")
    }
    try:
        loop = asyncio.get_event_loop()
        if loop.is_running():
            asyncio.run_coroutine_threadsafe(manager.broadcast(log_msg), loop)
    except Exception:
        pass

# Add sink to loguru
logger.add(websocket_sink, level="DEBUG")

class DownloadRequest(BaseModel):
    slug: str
    formats: List[str] = ["EPUB"]
    grouping: str = "tatca"
    tasks: int = 5
    skip_illustrations: bool = False
    output_folder: Optional[str] = None

active_tasks = {}

async def run_scrape_task(req: DownloadRequest, task_id: str):
    """Orchestrates the scraping task and sends progress via WebSocket."""
    try:
        await manager.broadcast({"type": "status", "task_id": task_id, "status": "Resolving story..."})
        
        session_state = load_session(".vvr_session.json")
        cookies = {}
        if session_state and 'cookies' in session_state:
            for c in session_state['cookies']:
                cookies[c['name']] = c['value']
        
        async with httpx.AsyncClient(headers=HEADERS, cookies=cookies) as client:
            story_info = await lay_thong_tin_truyen(client, req.slug)
            
        await manager.broadcast({"type": "info", "task_id": task_id, "title": story_info.title})
        
        story_url = f"{BASE_URL}/{req.slug}"
        chapter_data = await get_chapter_tree_list(story_url, output_file=f"chapters_{task_id}.json", session_state=session_state)

        if not chapter_data:
            if os.path.exists(f"chapters_{task_id}.json"):
                with open(f"chapters_{task_id}.json", "r", encoding="utf-8") as f:
                    chapter_data = json.load(f)
            else:
                raise Exception("Could not retrieve chapter list. Please check if the novel exists or try again.")

        if os.path.exists(f"chapters_{task_id}.json"):
            os.remove(f"chapters_{task_id}.json")

        selected_chaps = [c for v in chapter_data for c in v['chapters']]
        if req.skip_illustrations:
            selected_chaps = [c for c in selected_chaps if "Minh họa" not in c['title']]

        urls = [f"{BASE_URL}{c['url']}" for c in selected_chaps]
        total = len(urls)
        
        token = get_token_from_state(session_state)
        async with async_playwright() as p:
            browser = await p.chromium.launch(headless=True)
            scraped = {}
            semaphore = asyncio.Semaphore(req.tasks)
            
            async def process_url(url, idx):
                async with semaphore:
                    from .scraper_core import lay_chuong_httpx, lay_chuong_voi_hinh_anh
                    content = None
                    async with httpx.AsyncClient(headers=HEADERS, cookies=cookies, follow_redirects=True) as client:
                        content = await lay_chuong_httpx(client, url, token=token)
                    if not content:
                        content = await lay_chuong_voi_hinh_anh(browser, url, session_state=session_state)
                    
                    if content:
                        scraped[url] = content
                    
                    percent = int(((idx + 1) / total) * 100)
                    await manager.broadcast({
                        "type": "progress", 
                        "task_id": task_id, 
                        "percent": percent,
                        "msg": f"Downloaded {idx+1}/{total} chapters"
                    })

            tasks = [process_url(url, i) for i, url in enumerate(urls)]
            await asyncio.gather(*tasks)
            await browser.close()

        await manager.broadcast({"type": "status", "task_id": task_id, "status": "Exporting files..."})
        
        # Use provided output folder or default to sanitized title
        output_folder = req.output_folder or sanitize_filename(story_info.title)
        os.makedirs(output_folder, exist_ok=True)
        
        full_flat = []
        for url in urls:
            if url in scraped:
                full_flat.extend(scraped[url])
        
        for fmt in req.formats:
            fpath = os.path.join(output_folder, f"{sanitize_filename(story_info.title)}.{fmt.lower()}")
            if fmt == "PDF": await tao_file_pdf(full_flat, fpath, story_info.title)
            elif fmt == "EPUB": await tao_file_epub(fpath, story_info.title, story_info.author, [{'title': 'All Chapters', 'content': full_flat}], story_info.description, story_info.cover_path, story_info.genres)
            elif fmt == "HTML": await tao_file_html(full_flat, fpath, story_info.title)
            elif fmt == "MD": await tao_file_md(full_flat, fpath, story_info.title)
            elif fmt == "TXT": await tao_file_txt(full_flat, fpath, story_info.title)
            elif fmt == "MP3": await tao_file_mp3(full_flat, fpath, story_info.title)

        await manager.broadcast({"type": "complete", "task_id": task_id, "path": output_folder})
        logger.success(f"Task {task_id} completed: {story_info.title}")
    except Exception as e:
        import traceback
        logger.error(f"Task {task_id} failed: {e}")
        logger.error(traceback.format_exc())
        await manager.broadcast({"type": "error", "task_id": task_id, "error": str(e)})
    finally:
        if task_id in active_tasks:
            del active_tasks[task_id]

@app.get("/health")
async def health_check():
    return {"status": "ok"}

@app.get("/api/search")
async def search_novels(q: str = Query(..., min_length=3)):
    """Proxies search requests to the Valvrare Team API."""
    try:
        async with httpx.AsyncClient(timeout=10.0) as client:
            headers = {
                "User-Agent": HEADERS["User-Agent"],
                "Origin": BASE_URL,
                "Referer": f"{BASE_URL}/",
                "Accept": "application/json, text/plain, */*"
            }
            url = f"{SSR_API_URL}?title={q}"
            response = await client.get(url, headers=headers)
            response.raise_for_status()
            results = response.json()
            
            for item in results:
                title = item.get('title', '').strip()
                _id = item.get('_id', '')
                if title and _id:
                    item['slug'] = "truyen/" + normalize_vietnamese_url(title) + "-" + _id[-8:]
            return results
    except Exception as e:
        logger.error(f"Search error: {e}")
        return []

@app.get("/api/browse")
async def browse_folder():
    """Opens a native folder selection dialog on the host machine with fallbacks."""
    import os
    import subprocess

    # 1. Try zenity (common on Linux)
    try:
        proc = subprocess.run(
            ['zenity', '--file-selection', '--directory', '--title=Chọn thư mục đầu ra'],
            capture_output=True, text=True
        )
        if proc.returncode == 0 and proc.stdout.strip():
            return {"path": os.path.abspath(proc.stdout.strip())}
    except FileNotFoundError:
        pass

    # 2. Try kdialog (KDE fallback)
    try:
        proc = subprocess.run(
            ['kdialog', '--getexistingdirectory', '.', '--title', 'Chọn thư mục đầu ra'],
            capture_output=True, text=True
        )
        if proc.returncode == 0 and proc.stdout.strip():
            return {"path": os.path.abspath(proc.stdout.strip())}
    except FileNotFoundError:
        pass

    # 3. Try tkinter (System fallback)
    try:
        import tkinter as tk
        from tkinter import filedialog
        root = tk.Tk()
        root.withdraw()
        root.attributes('-topmost', True)
        folder_selected = filedialog.askdirectory()
        root.destroy()
        if folder_selected:
            return {"path": os.path.abspath(folder_selected)}
    except ImportError:
        logger.error("Tkinter not found. Please install it (e.g., sudo zypper install python3-tk)")
        return {"error": "Tính năng này yêu cầu 'python3-tk' hoặc 'zenity'. Vui lòng cài đặt qua package manager của bạn."}
    except Exception as e:
        logger.error(f"Error opening folder dialog: {e}")
        return {"error": str(e)}

    return {"path": None}

@app.post("/api/download")
async def download_novel(req: DownloadRequest):
    task_id = str(uuid.uuid4())[:8]
    active_tasks[task_id] = req
    await download_queue.add_task(req, task_id)
    return {"task_id": task_id}

@app.websocket("/ws/tasks")
async def websocket_endpoint(websocket: WebSocket):
    await manager.connect(websocket)
    try:
        while True:
            await websocket.receive_text()
    except WebSocketDisconnect:
        manager.disconnect(websocket)

# Mount static files
static_dir = os.path.join(os.path.dirname(__file__), "static")
if os.path.exists(static_dir):
    app.mount("/static", StaticFiles(directory=static_dir), name="static")

@app.get("/", response_class=HTMLResponse)
async def get_index():
    index_path = os.path.join(static_dir, "index.html")
    if os.path.exists(index_path):
        with open(index_path, "r", encoding="utf-8") as f:
            return f.read()
    return "<h1>Frontend not found</h1>"

@app.get("/style.css")
async def get_css():
    css_path = os.path.join(static_dir, "style.css")
    if os.path.exists(css_path):
        with open(css_path, "r", encoding="utf-8") as f:
            return HTMLResponse(content=f.read(), media_type="text/css")

@app.get("/app.js")
async def get_js():
    js_path = os.path.join(static_dir, "app.js")
    if os.path.exists(js_path):
        with open(js_path, "r", encoding="utf-8") as f:
            return HTMLResponse(content=f.read(), media_type="application/javascript")

async def run_web_server(host: str = "127.0.0.1", port: int = 8000, num_workers: int = 1):
    """Starts the Uvicorn server in the current event loop."""
    download_queue.num_workers = num_workers
    logger.info(f"Starting web server at http://{host}:{port} with {num_workers} workers")
    config = uvicorn.Config(app, host=host, port=port, log_level="info")
    server = uvicorn.Server(config)
    await server.serve()



"""
Utility functions for the web novel scraper.
"""
import sys
from typing import Dict, List, Optional
from loguru import logger
import re
import os

def configure_logger(verbose: bool = False):
    """Configures the loguru logger with appropriate levels."""
    logger.remove()  # Remove default handler
    
    # Custom format for nice output
    log_format = (
        "<green>{time:HH:mm:ss}</green> | "
        "<level>{level: <8}</level> | "
        "<cyan>{message}</cyan>"
    )
    
    # Show only INFO and above for non-verbose, DEBUG and above for verbose
    level = "DEBUG" if verbose else "INFO"
    
    logger.add(sys.stderr, level=level, format=log_format, colorize=True)

HEADERS = {
    "User-Agent": "Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/146.0.0.0 Safari/537.36",
    "Accept": "text/html,application/xhtml+xml,application/xml;q=0.9,image/avif,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.7",
    "Accept-Language": "en-US,en;q=0.9,vi;q=0.8,en-GB;q=0.7",
    "Referer": "https://valvrareteam.net/",
    "DNT": "1",
    "Upgrade-Insecure-Requests": "1",
    "Sec-Fetch-Dest": "document",
    "Sec-Fetch-Mode": "navigate",
    "Sec-Fetch-Site": "same-origin",
    "Sec-Fetch-User": "?1",
    "Sec-GPC": "1",
}


def sanitize_filename(name: str) -> str:
    """
    Sanitizes a string to be used as a valid filename or directory name.
    It removes illegal characters for most OSes.
    """
    if not name:
        return ""
    sanitized_name = re.sub(r'[\\/*?:\"<>|]', "", name)
    sanitized_name = sanitized_name.strip(' .')
    sanitized_name = re.sub(r'\s+', ' ', sanitized_name).strip()
    return sanitized_name


def create_folders_from_tree(tree_file: str, base_folder: str) -> None:
    """Creates directory structure based on a tree map file."""
    try:
        with open(tree_file, 'r', encoding='utf-8') as f:
            tree_data = f.readlines()
        for line in tree_data:
            folder_name = sanitize_filename(line.strip())
            if folder_name:
                folder_path = os.path.join(base_folder, folder_name)
                os.makedirs(folder_path, exist_ok=True)
    except FileNotFoundError:
        print(f"Lưu ý: file tree_map.txt không tồn tại, sẽ tạo thư mục gốc.")
        os.makedirs(base_folder, exist_ok=True)


# Vietnamese character normalization map for URL processing
VIETNAMESE_MAP = {
    'à': 'a', 'á': 'a', 'ả': 'a', 'ã': 'a', 'ạ': 'a', 'ă': 'a', 'ằ': 'a', 'ắ': 'a', 'ẳ': 'a', 'ẵ': 'a', 'ặ': 'a',
    'â': 'a', 'ầ': 'a', 'ấ': 'a', 'ẩ': 'a', 'ẫ': 'a', 'ậ': 'a', 'đ': 'd', 'è': 'e', 'é': 'e', 'ẻ': 'e', 'ẽ': 'e',
    'ẹ': 'e', 'ê': 'e', 'ề': 'e', 'ế': 'e', 'ể': 'e', 'ễ': 'e', 'ệ': 'e', 'ì': 'i', 'í': 'i', 'ỉ': 'i', 'ĩ': 'i',
    'ị': 'i', 'ò': 'o', 'ó': 'o', 'ỏ': 'o', 'õ': 'o', 'ọ': 'o', 'ô': 'o', 'ồ': 'o', 'ố': 'o', 'ổ': 'o', 'ỗ': 'o',
    'ộ': 'o', 'ơ': 'o', 'ờ': 'o', 'ớ': 'o', 'ở': 'o', 'ỡ': 'o', 'ợ': 'o', 'ù': 'u', 'ú': 'u', 'ủ': 'u', 'ũ': 'u',
    'ụ': 'u', 'ư': 'u', 'ừ': 'u', 'ứ': 'u', 'ử': 'u', 'ữ': 'u', 'ự': 'u', 'ỳ': 'y', 'ý': 'y', 'ỷ': 'y', 'ỹ': 'y', 'ỵ': 'y'
}


def normalize_vietnamese_url(text: str) -> str:
    """Normalize Vietnamese characters and strip ALL punctuation/special chars for URL matching."""
    if not text:
        return ""
    
    # 1. Lowercase
    text = text.lower()
    
    # 2. Replace Vietnamese characters with ASCII
    for key, value in VIETNAMESE_MAP.items():
        text = text.replace(key, value)
    
    # 3. Remove all non-alphanumeric characters except spaces
    # This removes commas, dots, tildes (~), colons, etc.
    text = re.sub(r'[^a-z0-9\s-]', '', text)
    
    # 4. Replace spaces with hyphens
    normalized = text.replace(" ", "-")
    
    # 5. Collapse multiple hyphens and strip leading/trailing hyphens
    normalized = re.sub(r'-+', '-', normalized).strip('-')
    
    return normalized


def get_token_from_state(state: Dict) -> Optional[str]:
    """
    Extracts the JWT token from Playwright's storage state (localStorage).
    Looks for common token keys like 'accessToken', 'token', or within 'auth-storage'.
    """
    if not state or 'origins' not in state:
        return None
    
    for origin in state['origins']:
        if "valvrareteam.net" in origin['origin']:
            ls = origin.get('localStorage', [])
            for item in ls:
                name = item.get('name', '')
                value = item.get('value', '')
                
                # Direct match
                if name in ['accessToken', 'token', 'jwt']:
                    return value
                
                # Check for auth-storage (often a JSON string)
                if name == 'auth-storage':
                    try:
                        import json
                        data = json.loads(value)
                        # Check for state.token or similar structures
                        if isinstance(data, dict):
                            state_data = data.get('state', {})
                            if isinstance(state_data, dict):
                                return state_data.get('token') or state_data.get('accessToken')
                    except Exception:
                        pass
    return None



"""
Tests for scraper.py - Main crawler functionality
"""
import pytest
import os
import sys
import tempfile
import json
import asyncio
from unittest.mock import AsyncMock, MagicMock, patch, mock_open

# Add parent directory to path for imports
sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(__file__))))

from vvr_scraper.utils import sanitize_filename, HEADERS
from vvr_scraper.exporter import (
    tao_file_epub,
    tao_file_pdf,
    tao_file_html,
    tao_file_md,
    tao_file_txt,
)


# =============================================================================
# TESTS FOR sanitize_filename
# =============================================================================

class TestSanitizeFilename:
    """Tests for the sanitize_filename function"""

    def test_removes_invalid_chars(self):
        """Test that invalid filename characters are removed"""
        assert sanitize_filename('test<file>*name?.txt') == 'testfilename.txt'
        assert sanitize_filename('file:name.txt') == 'filename.txt'
        assert sanitize_filename('test"quotes".txt') == 'testquotes.txt'
        assert sanitize_filename('file<angle>.txt') == 'fileangle.txt'

    def test_removes_backslash_and_pipe(self):
        """Test removal of backslash and pipe characters"""
        assert sanitize_filename('test\\backslash.txt') == 'testbackslash.txt'
        assert sanitize_filename('test|pipe.txt') == 'testpipe.txt'

    def test_strips_leading_trailing_spaces_dots(self):
        """Test that leading/trailing spaces and dots are stripped"""
        assert sanitize_filename('  filename.txt') == 'filename.txt'
        assert sanitize_filename('filename.txt  ') == 'filename.txt'
        assert sanitize_filename('.filename.txt') == 'filename.txt'
        assert sanitize_filename('filename.txt.') == 'filename.txt'

    def test_collapses_multiple_spaces(self):
        """Test that multiple spaces are collapsed to single space"""
        assert sanitize_filename('multiple   spaces') == 'multiple spaces'
        assert sanitize_filename('a    b    c') == 'a b c'

    def test_handles_empty_string(self):
        """Test handling of empty string"""
        assert sanitize_filename('') == ''

    def test_handles_none(self):
        """Test handling of None input"""
        assert sanitize_filename(None) == ''

    def test_handles_vietnamese_characters(self):
        """Test that Vietnamese characters are preserved"""
        assert sanitize_filename('truyện tranh') == 'truyện tranh'
        assert sanitize_filename('đầy đủ dấu') == 'đầy đủ dấu'

    def test_complex_filename(self):
        """Test with a complex realistic filename"""
        input_name = '  Chương 1: Mở Đầu <Phiên Bản Mới>*.txt  '
        expected = 'Chương 1 Mở Đầu Phiên Bản Mới.txt'
        assert sanitize_filename(input_name) == expected


# =============================================================================
# TESTS FOR EXPORT FUNCTIONS
# =============================================================================

@pytest.mark.asyncio
class TestTaoFileHtml:
    """Tests for HTML export function"""

    async def test_creates_valid_html_file(self, tmp_path):
        """Test that a valid HTML file is created"""
        content = [
            {'type': 'text', 'data': 'Paragraph 1'},
            {'type': 'text', 'data': 'Paragraph 2'},
        ]
        filepath = str(tmp_path / "test.html")

        await tao_file_html(content, filepath, "Test Chapter")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            html = f.read()

        assert '<!DOCTYPE html>' in html
        assert '<title>Test Chapter</title>' in html
        assert '<p>Paragraph 1</p>' in html
        assert '<p>Paragraph 2</p>' in html
        assert "lang='vi'" in html or 'lang="vi"' in html

    async def test_html_with_images(self, tmp_path):
        """Test HTML export with image content"""
        content = [
            {'type': 'text', 'data': 'Text before'},
            {'type': 'image', 'data': 'https://example.com/image.jpg'},
            {'type': 'text', 'data': 'Text after'},
        ]
        filepath = str(tmp_path / "test.html")

        await tao_file_html(content, filepath, "Test Chapter")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            html = f.read()

        assert 'src=\'https://example.com/image.jpg\'' in html or 'src="https://example.com/image.jpg"' in html

    async def test_html_file_unicode_content(self, tmp_path):
        """Test HTML export with Vietnamese unicode content"""
        content = [
            {'type': 'text', 'data': 'Đây là nội dung tiếng Việt'},
            {'type': 'text', 'data': 'Chữ ơ, ư, ê đầy đủ'},
        ]
        filepath = str(tmp_path / "test.html")

        await tao_file_html(content, filepath, "Chương Tiếng Việt")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            html = f.read()

        assert 'Đây là nội dung tiếng Việt' in html


@pytest.mark.asyncio
class TestTaoFileMd:
    """Tests for Markdown export function"""

    async def test_creates_valid_markdown_file(self, tmp_path):
        """Test that a valid Markdown file is created"""
        content = [
            {'type': 'text', 'data': 'Paragraph 1'},
            {'type': 'text', 'data': 'Paragraph 2'},
        ]
        filepath = str(tmp_path / "test.md")

        await tao_file_md(content, filepath, "Test Chapter")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            md = f.read()

        assert md.startswith('# Test Chapter')
        assert 'Paragraph 1' in md
        assert 'Paragraph 2' in md

    async def test_markdown_with_images(self, tmp_path):
        """Test Markdown export with image content"""
        content = [
            {'type': 'text', 'data': 'Text'},
            {'type': 'image', 'data': 'https://example.com/image.jpg'},
        ]
        filepath = str(tmp_path / "test.md")

        await tao_file_md(content, filepath, "Test Chapter")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            md = f.read()

        assert '![Minh họa](https://example.com/image.jpg)' in md


@pytest.mark.asyncio
class TestTaoFileTxt:
    """Tests for plain text export function"""

    async def test_creates_valid_txt_file(self, tmp_path):
        """Test that a valid text file is created"""
        content = [
            {'type': 'text', 'data': 'Line 1'},
            {'type': 'text', 'data': 'Line 2'},
        ]
        filepath = str(tmp_path / "test.txt")

        await tao_file_txt(content, filepath, "Test Chapter")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            txt = f.read()

        assert txt.startswith('Test Chapter')
        assert 'Line 1' in txt
        assert 'Line 2' in txt

    async def test_txt_with_images(self, tmp_path):
        """Test text export with image placeholders"""
        content = [
            {'type': 'text', 'data': 'Text'},
            {'type': 'image', 'data': 'https://example.com/image.jpg'},
        ]
        filepath = str(tmp_path / "test.txt")

        await tao_file_txt(content, filepath, "Test Chapter")

        assert os.path.exists(filepath)
        with open(filepath, 'r', encoding='utf-8') as f:
            txt = f.read()

        assert '[Ảnh: https://example.com/image.jpg]' in txt


@pytest.mark.asyncio
class TestTaoFileEpub:
    """Tests for EPUB export function"""

    async def test_creates_epub_file(self, tmp_path):
        """Test that an EPUB file is created"""
        chapters_data = [
            {
                'title': 'Chapter 1',
                'content': [
                    {'type': 'text', 'data': 'Hello World'},
                ]
            }
        ]
        filepath = str(tmp_path / "test.epub")

        await tao_file_epub(filepath, "Test Book", "Test Author", chapters_data, genres=["Action", "Fantasy"])

        assert os.path.exists(filepath)
        # EPUB files are ZIP archives, check basic structure
        import zipfile
        with zipfile.ZipFile(filepath, 'r') as zip_ref:
            names = zip_ref.namelist()
            assert any('chap_' in n for n in names)
            assert 'mimetype' in names

    async def test_epub_with_volume_structure(self, tmp_path):
        """Test EPUB export with volume/chapter hierarchy"""
        chapters_data = [
            {
                'volume': 'Volume 1',
                'chapters': [
                    {
                        'title': 'Chapter 1',
                        'content': [{'type': 'text', 'data': 'Content'}]
                    }
                ]
            }
        ]
        filepath = str(tmp_path / "test.epub")

        await tao_file_epub(filepath, "Test Book", "Test Author", chapters_data)

        assert os.path.exists(filepath)


@pytest.mark.asyncio
class TestTaoFilePdf:
    """Tests for PDF export function"""

    async def test_pdf_creation_basic(self, tmp_path):
        """Test basic PDF creation"""
        content = [
            {'type': 'text', 'data': 'Test content'},
        ]
        filepath = str(tmp_path / "test.pdf")

        # This test may fail if fonts are not available, which is expected
        await tao_file_pdf(content, filepath, "Test Chapter")

        # File should be created even if font fallback occurs
        assert os.path.exists(filepath)


# =============================================================================
# TESTS FOR HEADERS CONFIGURATION
# =============================================================================

class TestHeadersConfiguration:
    """Tests for HTTP headers configuration"""

    def test_headers_contains_required_fields(self):
        """Test that HEADERS contains all required fields"""
        assert 'User-Agent' in HEADERS
        assert 'Accept' in HEADERS
        assert 'Accept-Language' in HEADERS

    def test_user_agent_format(self):
        """Test User-Agent string format"""
        assert HEADERS['User-Agent'].startswith('Mozilla/5.0')
        assert 'Chrome' in HEADERS['User-Agent']


# =============================================================================
# INTEGRATION TESTS (Mocked)
# =============================================================================

class TestScrapingIntegration:
    """Integration tests with mocked external dependencies"""

    @pytest.mark.asyncio
    async def test_lay_thong_tin_truyen_structure(self):
        """Test story info scraping with mocked response"""
        from vvr_scraper.scraper_core import lay_thong_tin_truyen
        import httpx

        mock_html = """
        <html>
            <h1 class="rd-novel-title">Test Story</h1>
            <span class="rd-author-name">Author Name</span>
            <div class="rd-description-content">Story description</div>
            <span class="rd-genre-tag">Action</span>
            <span class="rd-genre-tag">Fantasy</span>
        </html>
        """

        mock_response = MagicMock()
        mock_response.text = mock_html
        mock_response.status_code = 200
        mock_response.raise_for_status = MagicMock()

        with patch('httpx.AsyncClient') as MockClient:
            instance = MockClient.return_value
            instance.get = AsyncMock(return_value=mock_response)
            instance.__aenter__ = AsyncMock(return_value=instance)
            instance.__aexit__ = AsyncMock(return_value=None)

            result = await lay_thong_tin_truyen(instance, "test-story")

            assert result.title == 'Test Story'
            assert result.author == 'Author Name'
            assert result.description == 'Story description'
            assert result.genres == ['Action', 'Fantasy']


# =============================================================================
# EDGE CASE TESTS
# =============================================================================

class TestEdgeCases:
    """Tests for edge cases and error handling"""

    @pytest.mark.asyncio
    async def test_empty_content_export(self, tmp_path):
        """Test exporting empty content"""
        content = []

        # HTML should still create valid file
        filepath = str(tmp_path / "empty.html")
        await tao_file_html(content, filepath, "Empty")
        assert os.path.exists(filepath)

    def test_very_long_filename(self):
        """Test sanitization of very long filename"""
        long_name = "a" * 500
        result = sanitize_filename(long_name)
        assert len(result) == len(long_name)  # Length preserved after sanitization

    def test_special_unicode_characters(self):
        """Test handling of special unicode characters"""
        test_cases = [
            ('file\u200bname.txt', 'filename.txt'),  # Zero-width space
            ('file\u00a0name.txt', 'file\u00a0name.txt'),  # Non-breaking space (kept)
        ]
        for input_name, expected in test_cases:
            result = sanitize_filename(input_name)
            # Just ensure no exception is raised

    @pytest.mark.asyncio
    async def test_mixed_content_types(self, tmp_path):
        """Test export with mixed text and multiple images"""
        content = [
            {'type': 'text', 'data': 'Intro'},
            {'type': 'image', 'data': 'https://example.com/img1.png'},
            {'type': 'text', 'data': 'Middle'},
            {'type': 'image', 'data': 'https://example.com/img2.gif'},
            {'type': 'text', 'data': 'End'},
        ]

        for fmt, func in [
            ('html', tao_file_html),
            ('md', tao_file_md),
            ('txt', tao_file_txt),
        ]:
            filepath = str(tmp_path / f"mixed.{fmt}")
            await func(content, filepath, "Test")
            assert os.path.exists(filepath)



"""
Core scraping functions for the web novel scraper.
"""
import asyncio
from typing import Dict, List, Optional, Any

import httpx
from playwright.async_api import Browser, async_playwright
from bs4 import BeautifulSoup

from loguru import logger

from .utils import HEADERS
from .models import StoryInfo, ContentItem


MAX_RETRIES = 2


async def lay_thong_tin_truyen(client: httpx.AsyncClient, ten_truyen: str, verbose: bool = False) -> StoryInfo:
    """
    Scrapes basic information about the story from its main page using httpx and BeautifulSoup.
    """
    url = f"https://valvrareteam.net/{ten_truyen}"
    logger.debug(f"Fetching story info from: {url}")
    
    response = await client.get(url, follow_redirects=True)
    response.raise_for_status()
    soup = BeautifulSoup(response.text, 'html.parser')

    title_element = soup.select_one("h1.rd-novel-title")
    title = title_element.get_text(strip=True) if title_element else "Unknown Title"
    
    # Clean up status suffixes from title
    for status in ["+Đang tiến hành", "+Hoàn thành", "+Tạm ngưng"]:
        if status in title:
            title = title.replace(status, "").strip()

    author_elements = soup.select("span.rd-author-name")
    authors = [author.get_text(strip=True) for author in author_elements]
    author = ", ".join(authors) if authors else "Unknown Author"

    description_element = soup.select_one("div.rd-description-content")
    description = description_element.get_text(strip=True) if description_element else "No Description"

    genre_elements = soup.select(".rd-genre-tag")
    genres = [genre.get_text(strip=True) for genre in genre_elements]

    cover_path = None
    image_url_element = soup.select_one("img.rd-cover-image")
    if image_url_element and 'src' in image_url_element.attrs:
        image_url = image_url_element['src']
        if image_url:
            try:
                response = await client.get(image_url, timeout=30.0)
                response.raise_for_status()
                cover_path = "cover.jpg"
                logger.info(f"Đã tải ảnh bìa: {cover_path}")
                with open(cover_path, "wb") as f:
                    f.write(response.content)
            except Exception as e:
                logger.warning(f"Không thể tải ảnh bìa: {e}")

    return StoryInfo(
        title=title,
        author=author,
        description=description,
        genres=genres,
        cover_path=cover_path
    )


async def lay_chuong_httpx(client: httpx.AsyncClient, url: str, verbose: bool = False, token: Optional[str] = None) -> Optional[List[ContentItem]]:
    """
    Scrapes a single chapter page using httpx and BeautifulSoup (Fast Mode).
    Uses the DigitalOcean SSR fallback for better reliability and speed.
    """
    fallback_url = url.replace("valvrareteam.net", "val-ssr-2kzit.ondigitalocean.app")
    logger.debug(f"Fast-scraping from: {fallback_url}")
    
    try:
        headers = client.headers.copy()
        if token:
            headers["Authorization"] = f"Bearer {token}"
            
        response = await client.get(fallback_url, follow_redirects=True, timeout=30.0, headers=headers)
        response.raise_for_status()
        soup = BeautifulSoup(response.text, 'html.parser')
        
        content_container = soup.select_one(".chapter-content")
        if not content_container:
            return None
            
        extracted_content: List[ContentItem] = []
        elements = content_container.select("p, img")
        
        for el in elements:
            if el.name == 'img':
                image_url = el.get('src')
                if image_url:
                    extracted_content.append(ContentItem(type='image', data=image_url))
            elif el.name == 'p':
                text = el.get_text(strip=True)
                if text:
                    extracted_content.append(ContentItem(type='text', data=text))
        
        return extracted_content if extracted_content else None
    except Exception as e:
        logger.debug(f"Fast-scrape failed for {fallback_url}: {e}")
        return None


async def lay_chuong_voi_hinh_anh(browser: Browser, url: str, session_state: Optional[Dict[str, Any]] = None, verbose: bool = False) -> Optional[List[ContentItem]]:
    """
    Scrapes a single chapter page for text and images using Playwright (Reliable Mode).
    """
    logger.debug(f"Playwright-scraping from: {url}")

    page = await browser.new_page(storage_state=session_state) if session_state else await browser.new_page()
    for attempt in range(MAX_RETRIES):
        try:
            await page.goto(url, wait_until='domcontentloaded', timeout=60000)
            content_selector = ".chapter-content p, .chapter-content img, .chapter-card p, .chapter-card img"
            await page.wait_for_selector(content_selector, timeout=30000)
            elements = page.locator(content_selector)
            extracted_content: List[ContentItem] = []
            for i in range(await elements.count()):
                element = elements.nth(i)
                tag_name = await element.evaluate('el => el.tagName')
                if tag_name == 'IMG':
                    image_url = await element.get_attribute('src')
                    if image_url:
                        extracted_content.append(ContentItem(type='image', data=image_url))
                elif tag_name == 'P':
                    text = await element.inner_text()
                    if text.strip():
                        extracted_content.append(ContentItem(type='text', data=text.strip()))
            await page.close()
            return extracted_content
        except Exception as e:
            if verbose:
                print(f"[DEBUG] Playwright attempt {attempt + 1} failed for {url}: {e}")
            if attempt < MAX_RETRIES - 1:
                await asyncio.sleep(2)
            else:
                if verbose:
                    print(f"[DEBUG] Playwright failed for {url} after {MAX_RETRIES} attempts.")

    await page.close()
    return None


async def scrape_chapters(
    browser: Browser,
    urls: List[str],
    concurrent_tasks: int = 5,
    skipped_urls: Optional[List[str]] = None,
    session_state: Optional[Dict[str, Any]] = None,
    verbose: bool = False,
    token: Optional[str] = None
) -> Dict[str, List[ContentItem]]:
    """
    Scrape multiple chapters concurrently.
    Uses a hybrid approach:
    1. Try HTTTPX (Fast) first.
    2. Fallback to Playwright (Reliable) if HTTPX fails or returns empty.
    """
    semaphore = asyncio.Semaphore(concurrent_tasks)
    scraped_content: Dict[str, List[ContentItem]] = {}
    if skipped_urls is None:
        skipped_urls_local: List[str] = []
        skipped_urls = skipped_urls_local
    else:
        skipped_urls_local = skipped_urls

    # Convert session_state to httpx cookies
    cookies = {}
    if session_state and 'cookies' in session_state:
        for c in session_state['cookies']:
            cookies[c['name']] = c['value']

    async def process_url(browser: Browser, url: str) -> None:
        async with semaphore:
            content = None
            # 1. Try Fast Mode (HTTPX)
            async with httpx.AsyncClient(headers=HEADERS, cookies=cookies, follow_redirects=True) as client:
                content = await lay_chuong_httpx(client, url, verbose=verbose, token=token)
            
            # 2. Try Reliable Mode (Playwright) if Fast Mode failed
            if not content:
                if verbose:
                    print(f"[*] Fast-scrape failed or empty for {url}. Falling back to Playwright...")
                content = await lay_chuong_voi_hinh_anh(browser, url, session_state=session_state, verbose=verbose)
            
            if content:
                scraped_content[url] = content
            else:
                skipped_urls.append(url)
                print(f"(!) Thất bại sau cả 2 phương thức: {url}")

    tasks = [process_url(browser, url) for url in urls]
    await asyncio.gather(*tasks)
    return scraped_content


import json
import os
from typing import Optional, Dict, Any
from playwright.async_api import async_playwright

def save_session(state: Dict[str, Any], file_path: str):
    """Saves the browser storage state (cookies and local storage) to a file."""
    with open(file_path, 'w', encoding='utf-8') as f:
        json.dump(state, f, ensure_ascii=False, indent=2)

def load_session(file_path: str) -> Optional[Dict[str, Any]]:
    """Loads the browser storage state from a file."""
    if not os.path.exists(file_path):
        return None
    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            return json.load(f)
    except Exception:
        return None

async def capture_session(url: str) -> Dict[str, Any]:
    """
    Launches a non-headless browser for the user to login or bypass Cloudflare.
    Returns the storage state upon user confirmation.
    """
    async with async_playwright() as p:
        # Launch non-headless browser
        browser = await p.chromium.launch(headless=False)
        context = await browser.new_context()
        page = await context.new_page()
        
        await page.goto(url)
        
        print("\n" + "="*60)
        print("VUI LÒNG ĐĂNG NHẬP HOẶC GIẢI CLOUDFLARE TRONG TRÌNH DUYỆT.")
        print("Sau khi hoàn tất và thấy nội dung truyện ĐÃ MỞ KHÓA, hãy quay lại đây.")
        print("Nhấn phím ENTER trong terminal này để tiếp tục...")
        print("="*60 + "\n")
        
        # Use input() which is more standard for this
        input()
        
        state = await context.storage_state()
        print(f"Đã bắt được session với {len(state.get('cookies', []))} cookies.")
        await browser.close()
        return state
